@@ -145,7 +145,6 @@ typedef struct {
 /* commands for use by keybindings */
 static void create(const char *args[]);
 static void copymode(const char *args[]);
-static void escapekey(const char *args[]);
 static void focusn(const char *args[]);
 static void focusnext(const char *args[]);
 static void focusnextnm(const char *args[]);
@@ -157,6 +156,7 @@ static void paste(const char *args[]);
 static void quit(const char *args[]);
 static void redraw(const char *args[]);
 static void scrollback(const char *args[]);
+static void send(const char *args[]);
 static void setlayout(const char *args[]);
 static void setmfact(const char *args[]);
 static void startup(const char *args[]);
@@ -218,20 +218,22 @@ isarrange(void (*func)()) {
 	return func == layout->arrange;
 }
 
-static void
-clear_workspace() {
-	for (unsigned int y = 0; y < wah; y++)
-		mvhline(way + y, 0, ' ', waw);
-	wnoutrefresh(stdscr);
+static bool
+is_content_visible(Client *c) {
+	if (!c)
+		return false;
+	if (isarrange(fullscreen))
+		return sel == c;
+	return !c->minimized;
 }
 
 static void
 drawbar() {
 	wchar_t wbuf[sizeof bar.text];
-	int w, maxwidth = screen.w - 2;
-	if (bar.pos == BAR_OFF || !bar.text[0])
+	int x, y, w, maxwidth = screen.w - 2;
+	if (bar.pos == BAR_OFF)
 		return;
-	curs_set(0);
+	getyx(stdscr, y, x);
 	attrset(BAR_ATTR);
 	mvaddch(bar.y, 0, '[');
 	if (mbstowcs(wbuf, bar.text, sizeof bar.text) == (size_t)-1)
@@ -249,9 +251,8 @@ drawbar() {
 	}
 	mvaddch(bar.y, screen.w - 1, ']');
 	attrset(NORMAL_ATTR);
-	if (sel)
-		curs_set(vt_cursor(sel->term));
-	refresh();
+	move(y, x);
+	wnoutrefresh(stdscr);
 }
 
 static void
@@ -261,7 +262,6 @@ draw_border(Client *c) {
 
 	wattrset(c->window, (sel == c || (runinall && !c->minimized)) ? SELECTED_ATTR : NORMAL_ATTR);
 	getyx(c->window, y, x);
-	curs_set(0);
 	mvwhline(c->window, 0, 0, ACS_HLINE, c->w);
 	maxlen = c->w - (2 + sstrlen(TITLE) - sstrlen("%s%sd")  + sstrlen(SEPARATOR) + 2);
 	if (maxlen < 0)
@@ -278,59 +278,50 @@ draw_border(Client *c) {
 	if (t)
 		c->title[maxlen] = t;
 	wmove(c->window, y, x);
-	if (!c->minimized)
-		curs_set(vt_cursor(c->term));
 }
 
 static void
 draw_content(Client *c) {
-	if (!c->minimized || isarrange(fullscreen)) {
-		vt_draw(c->term, c->window, 1, 0);
-		if (c != sel)
-			curs_set(0);
-	}
+	vt_draw(c->term, c->window, 1, 0);
 }
 
 static void
 draw(Client *c) {
-	draw_content(c);
-	draw_border(c);
-	wrefresh(c->window);
+	if (is_content_visible(c)) {
+		redrawwin(c->window);
+		draw_content(c);
+	}
+	if (!isarrange(fullscreen) || sel == c)
+		draw_border(c);
+	wnoutrefresh(c->window);
 }
 
 static void
-draw_all(bool border) {
+draw_all() {
 	Client *c;
-	curs_set(0);
-	for (c = clients; c; c = c->next) {
-		redrawwin(c->window);
-		if (c == sel)
-			continue;
-		draw_content(c);
-		if (border)
-			draw_border(c);
-		wnoutrefresh(c->window);
+	if (!isarrange(fullscreen)) {
+		for (c = clients; c; c = c->next) {
+			if (c == sel)
+				continue;
+			draw(c);
+		}
 	}
 	/* as a last step the selected window is redrawn,
 	 * this has the effect that the cursor position is
 	 * accurate
 	 */
-	refresh();
-	if (sel) {
-		draw_content(sel);
-		if (border)
-			draw_border(sel);
-		wrefresh(sel->window);
-	}
+	if (sel)
+		draw(sel);
 }
 
 static void
 arrange() {
-	clear_workspace();
+	erase();
+	drawbar();
 	attrset(NORMAL_ATTR);
 	layout->arrange();
 	wnoutrefresh(stdscr);
-	draw_all(true);
+	draw_all();
 }
 
 static void
@@ -379,10 +370,10 @@ detach(Client *c) {
 
 static void
 settitle(Client *c) {
-	char *t = title;
+	char *term, *t = title;
 	if (!t && sel == c && *c->title)
 		t = c->title;
-	if (t)
+	if (t && (term = getenv("TERM")) && !strstr(term, "linux"))
 		printf("\033]0;%s\007", t);
 }
 
@@ -393,14 +384,17 @@ focus(Client *c) {
 		return;
 	sel = c;
 	settitle(c);
-	if (tmp) {
+	if (tmp && !isarrange(fullscreen)) {
 		draw_border(tmp);
-		wrefresh(tmp->window);
+		wnoutrefresh(tmp->window);
 	}
-	if (isarrange(fullscreen))
-		redrawwin(c->window);
-	draw_border(c);
-	wrefresh(c->window);
+	if (isarrange(fullscreen)) {
+		draw(c);
+	} else {
+		draw_border(c);
+		wnoutrefresh(c->window);
+	}
+	curs_set(!c->minimized && vt_cursor(c->term));
 }
 
 static void
@@ -431,7 +425,8 @@ term_event_handler(Vt *term, int event, void *event_data) {
 			strncpy(c->title, event_data, sizeof(c->title) - 1);
 		c->title[event_data ? sizeof(c->title) - 1 : 0] = '\0';
 		settitle(c);
-		draw_border(c);
+		if (!isarrange(fullscreen) || sel == c)
+			draw_border(c);
 		applycolorrules(c);
 		break;
 	case VT_EVENT_COPY_TEXT:
@@ -568,13 +563,10 @@ resize_screen() {
 
 	resizeterm(screen.h, screen.w);
 	wresize(stdscr, screen.h, screen.w);
-	wrefresh(curscr);
-	refresh();
 
 	waw = screen.w;
 	wah = screen.h;
 	updatebarpos();
-	drawbar();
 	arrange();
 }
 
@@ -613,7 +605,7 @@ keypress(int code) {
 	}
 
 	for (c = runinall ? clients : sel; c; c = c->next) {
-		if (!c->minimized || isarrange(fullscreen)) {
+		if (is_content_visible(c)) {
 			if (code == '\e')
 				vt_write(c->term, buf, len);
 			else
@@ -676,7 +668,7 @@ destroy(Client *c) {
 			sel = NULL;
 	}
 	werase(c->window);
-	wrefresh(c->window);
+	wnoutrefresh(c->window);
 	vt_destroy(c->term);
 	delwin(c->window);
 	if (!clients && countof(actions)) {
@@ -777,15 +769,6 @@ copymode(const char *args[]) {
 	}
 }
 
-static void
-escapekey(const char *args[]) {
-	int key;
-	if ((key = getch()) >= 0) {
-		debug("escaping key `%c'\n", key);
-		keypress(CTRL(key));
-	}
-}
-
 static void
 focusn(const char *args[]) {
 	Client *c;
@@ -910,11 +893,18 @@ quit(const char *args[]) {
 
 static void
 redraw(const char *args[]) {
-	for (Client *c = clients; c; c = c->next)
-		vt_dirty(c->term);
-	wrefresh(curscr);
+	for (Client *c = clients; c; c = c->next) {
+		if (!c->minimized) {
+			vt_dirty(c->term);
+			wclear(c->window);
+			wnoutrefresh(c->window);
+		}
+	}
+	wclear(stdscr);
+	wnoutrefresh(stdscr);
+	doupdate();
 	resize_screen();
-	draw_all(true);
+	draw_all();
 }
 
 static void
@@ -929,6 +919,12 @@ scrollback(const char *args[]) {
 	draw(sel);
 }
 
+static void
+send(const char *args[]) {
+	if (sel && args && args[0])
+		vt_write(sel->term, args[0], strlen(args[0]));
+}
+
 static void
 setlayout(const char *args[]) {
 	unsigned int i;
@@ -983,7 +979,6 @@ togglebar(const char *args[]) {
 		bar.pos = BAR_OFF;
 	updatebarpos();
 	arrange();
-	drawbar();
 }
 
 static void
@@ -1040,7 +1035,7 @@ togglemouse(const char *args[]) {
 static void
 togglerunall(const char *args[]) {
 	runinall = !runinall;
-	draw_all(true);
+	draw_all();
 }
 
 static void
@@ -1232,11 +1227,12 @@ handle_statusbar() {
 			bar.fd = -1;
 			break;
 		default:
-			bar.text[r] = '\0'; p = bar.text + strlen(bar.text) - 1;
+			bar.text[r] = '\0';
+			p = bar.text + r - 1;
 			for (; p >= bar.text && *p == '\n'; *p-- = '\0');
 			for (; p >= bar.text && *p != '\n'; --p);
-			if (p > bar.text)
-				strncpy(bar.text, p + 1, sizeof bar.text);
+			if (p >= bar.text)
+				memmove(bar.text, p + 1, strlen(p));
 			drawbar();
 	}
 }
@@ -1280,7 +1276,7 @@ parse_args(int argc, char *argv[]) {
 		set_escdelay(100);
 	for (arg = 1; arg < argc; arg++) {
 		if (argv[arg][0] != '-') {
-			const char *args[] = { argv[arg], NULL };
+			const char *args[] = { argv[arg], NULL, NULL };
 			if (!init) {
 				setup();
 				init = true;
@@ -1339,6 +1335,8 @@ parse_args(int argc, char *argv[]) {
 
 int
 main(int argc, char *argv[]) {
+	int mod = ERR;
+
 	if (!parse_args(argc, argv)) {
 		setup();
 		startup(NULL);
@@ -1378,6 +1376,7 @@ main(int argc, char *argv[]) {
 			nfds = max(nfds, c->pty);
 			c = c->next;
 		}
+		doupdate();
 		r = select(nfds + 1, &rd, NULL, NULL, NULL);
 
 		if (r == -1 && errno == EINTR)
@@ -1392,17 +1391,14 @@ main(int argc, char *argv[]) {
 			int code = getch();
 			Key *key;
 			if (code >= 0) {
-				if (code == KEY_MOUSE) {
+				if (mod >= 0) {
+					if ((key = keybinding(mod, code)))
+						key->action.cmd(key->action.args);
+					mod = ERR;
+				} else if (code == KEY_MOUSE) {
 					handle_mouse();
 				} else if (is_modifier(code)) {
-					int mod = code;
-					code = getch();
-					if (code >= 0) {
-						if (code == mod)
-							keypress(code);
-						else if ((key = keybinding(mod, code)))
-							key->action.cmd(key->action.args);
-					}
+					mod = code;
 				} else if ((key = keybinding(0, code))) {
 					key->action.cmd(key->action.args);
 				} else if (sel && vt_copymode(sel->term)) {
@@ -1431,20 +1427,19 @@ main(int argc, char *argv[]) {
 					c = t;
 					continue;
 				}
-				if (c != sel) {
+				if (c != sel && is_content_visible(c)) {
 					draw_content(c);
-					if (!isarrange(fullscreen))
-						wnoutrefresh(c->window);
+					wnoutrefresh(c->window);
 				}
 			}
 			c = c->next;
 		}
 
-		if (sel) {
+		if (is_content_visible(sel)) {
 			draw_content(sel);
+			curs_set(vt_cursor(sel->term));
 			wnoutrefresh(sel->window);
 		}
-		doupdate();
 	}
 
 	cleanup();
