@@ -21,6 +21,7 @@
 #include <errno.h>
 #include <fcntl.h>
 #include <langinfo.h>
+#include <limits.h>
 #include <signal.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -197,6 +198,7 @@ static const char *keytable[KEY_MAX+1] = {
 	[KEY_NPAGE]     = "\e[6~",
 	[KEY_HOME]      = "\e[7~",
 	[KEY_END]       = "\e[8~",
+	[KEY_BTAB]      = "\e[Z",
 	[KEY_SUSPEND]   = "\x1A",  /* Ctrl+Z gets mapped to this */
 	[KEY_F(1)]      = "\e[11~",
 	[KEY_F(2)]      = "\e[12~",
@@ -221,6 +223,9 @@ static const char *keytable[KEY_MAX+1] = {
 	[KEY_F(21)]     = "\e[33~",
 	[KEY_F(22)]     = "\e[34~",
 	[KEY_RESIZE]    = "",
+#ifdef KEY_EVENT
+	[KEY_EVENT]     = "",
+#endif
 };
 
 static void puttab(Vt *t, int count);
@@ -711,6 +716,42 @@ static void interpret_csi_decstbm(Vt *t, int param[], int pcount)
 	b->curs_col = 0;
 }
 
+static void interpret_csi_mode(Vt *t, int param[], int pcount, bool set)
+{
+	for (int i = 0; i < pcount; i++) {
+		switch (param[i]) {
+		case 4: /* insert/replace mode */
+			t->insert = set;
+			break;
+		}
+	}
+}
+
+static void interpret_csi_priv_mode(Vt *t, int param[], int pcount, bool set)
+{
+	for (int i = 0; i < pcount; i++) {
+		switch (param[i]) {
+		case 1: /* set application/normal cursor key mode (DECCKM) */
+			t->curskeymode = set;
+			break;
+		case 6: /* set origin to relative/absolute (DECOM) */
+			t->relposmode = set;
+			break;
+		case 25: /* make cursor visible/invisible (DECCM) */
+			t->curshid = !set;
+			break;
+		case 47: /* use alternate/normal screen buffer */
+			vt_copymode_leave(t);
+			t->buffer = set ? &t->buffer_alternate : &t->buffer_normal;
+			vt_dirty(t);
+			break;
+		case 1000: /* enable/disable normal mouse tracking */
+			t->mousetrack = set;
+			break;
+		}
+	}
+}
+
 static void interpret_csi(Vt *t)
 {
 	static int csiparam[BUFSIZ];
@@ -736,67 +777,29 @@ static void interpret_csi(Vt *t)
 	}
 
 	if (t->ebuf[1] == '?') {
-		if (verb == 'h') { /* DEC Private Mode Set (DECSET) */
-			switch (csiparam[0]) {
-			case 1: /* set ANSI cursor (application) key mode (DECCKM) */
-				t->curskeymode = true;
-				break;
-			case 6: /* set origin to relative (DECOM) */
-				t->relposmode = true;
-				break;
-			case 25: /* make cursor visible (DECCM) */
-				t->curshid = false;
-				break;
-			case 47: /* use alternate screen buffer */
-				vt_copymode_leave(t);
-				t->buffer = &t->buffer_alternate;
-				vt_dirty(t);
-				break;
-			case 1000: /* enable normal mouse tracking */
-				t->mousetrack = true;
-				break;
-			}
-		} else if (verb == 'l') { /* DEC Private Mode Reset (DECRST) */
-			switch (csiparam[0]) {
-			case 1: /* reset ANSI cursor (normal) key mode (DECCKM) */
-				t->curskeymode = false;
-				break;
-			case 6: /* set origin to absolute (DECOM) */
-				t->relposmode = false;
-				break;
-			case 25: /* make cursor visible (DECCM) */
-				t->curshid = true;
-				break;
-			case 47: /* use normal screen buffer */
-				vt_copymode_leave(t);
-				t->buffer = &t->buffer_normal;
-				vt_dirty(t);
-				break;
-			case 1000: /* disable normal mouse tracking */
-				t->mousetrack = false;
-				break;
-			}
+		switch (verb) {
+		case 'h':
+		case 'l': /* private set/reset mode */
+			interpret_csi_priv_mode(t, csiparam, param_count, verb == 'h');
+			break;
 		}
+		return;
 	}
 
 	/* delegate handling depending on command character (verb) */
 	switch (verb) {
 	case 'h':
-		if (param_count == 1 && csiparam[0] == 4) /* insert mode */
-			t->insert = true;
+	case 'l': /* set/reset mode */
+		interpret_csi_mode(t, csiparam, param_count, verb == 'h');
 		break;
-	case 'l':
-		if (param_count == 1 && csiparam[0] == 4) /* replace mode */
-			t->insert = false;
-		break;
-	case 'm': /* it's a 'set attribute' sequence */
+	case 'm': /* set attribute */
 		interpret_csi_sgr(t, csiparam, param_count);
 		break;
-	case 'J': /* it's an 'erase display' sequence */
+	case 'J': /* erase display */
 		interpret_csi_ed(t, csiparam, param_count);
 		break;
 	case 'H':
-	case 'f': /* it's a 'move cursor' sequence */
+	case 'f': /* move cursor */
 		interpret_csi_cup(t, csiparam, param_count);
 		break;
 	case 'A':
@@ -809,8 +812,7 @@ static void interpret_csi(Vt *t)
 	case 'e':
 	case 'a':
 	case 'd':
-	case '`':
-		/* it is a 'relative move' */
+	case '`': /* relative move */
 		interpret_csi_c(t, verb, csiparam, param_count);
 		break;
 	case 'K': /* erase line */
@@ -1151,7 +1153,7 @@ static void put_wc(Vt *t, wchar_t wc)
 			Cell *src = b->curs_row->cells + b->curs_col;
 			Cell *dest = src + width;
 			size_t len = b->cols - b->curs_col - width;
-			memmove(dest, src, len);
+			memmove(dest, src, len * sizeof *dest);
 		}
 
 		b->curs_row->cells[b->curs_col] = blank_cell;
@@ -1386,6 +1388,7 @@ void vt_destroy(Vt *t)
 	buffer_free(&t->buffer_normal);
 	buffer_free(&t->buffer_alternate);
 	cmdline_free(t->cmdline);
+	close(t->pty);
 	free(t);
 }
 
@@ -1455,7 +1458,6 @@ void vt_draw(Vt *t, WINDOW * win, int srow, int scol)
 	int sel_col_start, sel_col_end;
 
 	copymode_get_selection_boundry(t, &sel_row_start, &sel_col_start, &sel_row_end, &sel_col_end, true);
-	curs_set(0);
 
 	for (int i = 0; i < b->rows; i++) {
 		Row *row = b->lines + i;
@@ -1526,8 +1528,6 @@ void vt_draw(Vt *t, WINDOW * win, int srow, int scol)
 		} else
 			wmove(win, srow + b->rows - 1, 1);
 	}
-
-	curs_set(vt_cursor(t));
 }
 
 void vt_scroll(Vt *t, int rows)
@@ -1616,9 +1616,12 @@ int vt_write(Vt *t, const char *buf, int len)
 
 	while (len > 0) {
 		int res = write(t->pty, buf, len);
-		if (res < 0 && errno != EAGAIN && errno != EINTR)
-			return -1;
-
+		if (res < 0) {
+			if (errno != EAGAIN && errno != EINTR)
+				return -1;
+			else
+				continue;
+		}
 		buf += res;
 		len -= res;
 	}
@@ -1636,11 +1639,9 @@ static void send_curs(Vt *t)
 
 void vt_keypress(Vt *t, int keycode)
 {
-	char c = (char)keycode;
-
 	vt_noscroll(t);
 
-	if (keycode >= 0 && keycode < KEY_MAX && keytable[keycode]) {
+	if (keycode >= 0 && keycode <= KEY_MAX && keytable[keycode]) {
 		switch (keycode) {
 		case KEY_UP:
 		case KEY_DOWN:
@@ -1653,8 +1654,13 @@ void vt_keypress(Vt *t, int keycode)
 		default:
 			vt_write(t, keytable[keycode], strlen(keytable[keycode]));
 		}
-	} else {
+	} else if (keycode <= UCHAR_MAX) {
+		char c = keycode;
 		vt_write(t, &c, 1);
+	} else {
+#ifndef NDEBUG
+		fprintf(stderr, "unhandled key %#o\n", keycode);
+#endif
 	}
 }
 
@@ -1873,6 +1879,14 @@ static void cmdline_keypress(Cmdline *c, int keycode)
 	ssize_t len;
 	size_t n = (c->end - c->cursor) * sizeof(wchar_t);
 	switch (keycode) {
+	case KEY_DC:
+		if (c->cursor == c->end) /* n == 0 */
+			break;
+		memmove(c->cursor, c->cursor + 1, n - sizeof(wchar_t));
+		c->end--;
+		*c->end = L'\0';
+		cmdline_adjust_cursor_pos(c);
+		break;
 	case KEY_BACKSPACE:
 		if (c->cursor == c->buf)
 			break;
